import os
import torch
import transformers
import bitsandbytes as bnb

from dataclasses import dataclass, field
from typing import Optional, Dict
from transformers import BitsAndBytesConfig
from peft import prepare_model_for_kbit_training, LoraConfig, get_peft_model
from transformers import Trainer

from llamatune.utils import _get_compute_dtype, print_trainable_parameters
from llamatune.telemetry.send_event import capture_event



class LlamaEngine:

    def __init__(self, model_name,  training_config):
        if not training_config.training_recipe in ["lora", "full_training"]:
            raise Exception(f"{training_config.training_recipe} is not a valid training recipe. Please choose either \"lora\" or \"full_training\"")
        
        self.config = training_config
        self.training_recipe = training_config.training_recipe
        self.model_name = model_name

    
    def train(self, data_module):
        print("config", self.config)

        training_args = self.construct_training_arguments()

        trainer = Trainer(
            model=self.model,
            tokenizer=self.tokenizer,
            train_dataset=data_module.train_dataset,
            data_collator=data_module.data_collator,
            args=training_args
        )

        trainer.train()
        capture_event("end-training", {})


    def prepare_model_for_training(self):
        capture_event("start-training", {})

        if self.config.training_recipe == "lora":
            self.model = transformers.AutoModelForCausalLM.from_pretrained(
                self.model_name,
                load_in_4bit=self.config.bits == 4,
                load_in_8bit=self.config.bits == 8,
                device_map=self._get_device_map(),
                torch_dtype=self.config.compute_dtype,
                trust_remote_code=self.config.trust_remote_code,
                use_auth_token=self.config.use_auth_token,
                quantization_config=BitsAndBytesConfig(
                    load_in_4bit=self.config.bits == 4,
                    load_in_8bit=self.config.bits == 8,
                    llm_int8_threshold=6.0,
                    llm_int8_has_fp16_weight=False,
                    bnb_4bit_compute_dtype=self.config.compute_dtype,
                    bnb_4bit_use_double_quant=self.config.double_quant,
                    bnb_4bit_quant_type=self.config.quant_type,
                    trust_remote_code=self.config.trust_remote_code
                )
            )

        elif self.config.training_recipe == "full_training":
            self.model = transformers.AutoModelForCausalLM.from_pretrained(
                self.model_name,
                device_map=self._get_device_map(),
                torch_dtype=self.config.compute_dtype,
                trust_remote_code=self.config.trust_remote_code,
                use_auth_token=self.config.use_auth_token
            )


        self.tokenizer = transformers.AutoTokenizer.from_pretrained(self.model_name, padding_side="right", use_fast=self.config.use_fast, tokenizer_type=self.config.tokenizer_type, trust_remote_code=self.config.trust_remote_code, use_auth_token=self.config.use_auth_token)

        self._smart_tokenizer_and_embedding_resize()

        if self.training_recipe == "lora":
            self.model = prepare_model_for_kbit_training(self.model, use_gradient_checkpointing=self.config.gradient_checkpointing)

        if self.config.gradient_checkpointing:
            self.model.gradient_checkpointing_enable()


        if self.training_recipe == "lora":
            modules = self._find_all_linear_names()

            lora_config = LoraConfig(
                r=self.config.lora_r,
                lora_alpha=self.config.lora_alpha,
                target_modules=modules,
                lora_dropout=self.config.lora_dropout,
                bias="none",
                task_type="CAUSAL_LM",
            )
            
            self.model = get_peft_model(self.model, lora_config)

        print("Model ready for training!")
        print_trainable_parameters(self.model)


    def _get_device_map(self):
        device_map = "auto"

        if os.environ.get("LOCAL_RANK") is not None:
            device_map = {'': int(os.environ.get('LOCAL_RANK', '0'))}
        
        return device_map
    
    def construct_training_arguments(self):

        args=transformers.TrainingArguments(
                output_dir = self.config.output_dir,
                optim = self.config.optim,
                per_device_train_batch_size = self.config.batch_size,
                gradient_accumulation_steps = self.config.gradient_accumulation_steps,
                num_train_epochs = self.config.n_epochs,
                weight_decay = self.config.weight_decay,
                learning_rate = self.config.learning_rate,
                max_grad_norm = self.config.max_grad_norm,
                gradient_checkpointing = self.config.gradient_checkpointing,
                do_train = self.config.do_train,
                lr_scheduler_type = self.config.lr_scheduler_type,
                warmup_ratio = self.config.warmup_ratio,
                logging_steps = self.config.logging_steps,
                group_by_length = self.config.group_by_length,
                save_strategy = self.config.save_strategy,
            )

        return args


    def _find_all_linear_names(self):
        cls = bnb.nn.Linear4bit if self.config.bits == 4 else (bnb.nn.Linear8bitLt if self.config.bits == 8 else torch.nn.Linear)
        lora_module_names = set()
        for name, module in self.model.named_modules():
            if isinstance(module, cls):
                names = name.split('.')
                lora_module_names.add(names[0] if len(names) == 1 else names[-1])

        if "lm_head" in lora_module_names: # needed for 16-bit
            lora_module_names.remove("lm_head")

        return list(lora_module_names)
    

    def _smart_tokenizer_and_embedding_resize(self):
        if self.tokenizer.pad_token is None:
            num_new_tokens = self.tokenizer.add_special_tokens(dict(pad_token="[PAD]"))
            self.model.resize_token_embeddings(len(self.tokenizer))

            if num_new_tokens > 0:
                input_embeddings = self.model.get_input_embeddings().weight.data
                output_embeddings = self.model.get_output_embeddings().weight.data

                input_embeddings_avg = input_embeddings[:-num_new_tokens].mean(dim=0, keepdim=True)
                output_embeddings_avg = output_embeddings[:-num_new_tokens].mean(dim=0, keepdim=True)

                input_embeddings[-num_new_tokens:] = input_embeddings_avg
                output_embeddings[-num_new_tokens:] = output_embeddings_avg
        



import transformers
from typing import Dict
from dataclasses import dataclass, field
import torch

from llamatune.model_engines.llama_model_engine import LlamaEngine
from llamatune.data.chat_data_module import ChatDataModule


class ChatTrainer:
    def __init__(self, training_config: dataclass):
        self.training_config = training_config
        self.model_engine : LlamaEngine = LlamaEngine(training_config.model_name, training_config=training_config)
        self.data_path = training_config.data_path

    def train(self):
        self.model_engine.prepare_model_for_training()

        self.data_module = ChatDataModule(
            tokenizer=self.model_engine.tokenizer,
            data_path_train=self.data_path,
            max_tokens=self.model_engine.config.max_tokens
        )

        self.model_engine.train(data_module=self.data_module)
        



import torch
import transformers
import io
import json
import logging

from dataclasses import dataclass
from typing import Dict, Sequence
from torch.utils.data import Dataset
from llamatune.data.data_module import DataModule


SYS_PREFIX = "<<SYS>>\n"
SYS_POSTFIX = "\n<</SYS>>\n\n"
INST_PREFIX = "<s>[INST] "
INST_POSTFIX = " "
OUTPUT_PREFIX = "[/INST] "
OUTPUT_POSTFIX = "</s>"


def _make_r_io_base(f, mode: str):
    if not isinstance(f, io.IOBase):
        f = open(f, mode=mode)
    return f


def jload(f, mode="r"):
    """Load a .json file into a dictionary."""
    f = _make_r_io_base(f, mode)
    jdict = json.load(f)
    f.close()
    return jdict


class ChatDataset(Dataset):
    def __init__(self, data_path: str, tokenizer: transformers.AutoTokenizer, max_tokens=None):
        super(ChatDataset, self).__init__()
        logging.warning("Loading data...")
        conversations = jload(data_path)

        logging.warning("Tokenizing inputs... This may take some time...")
        data_dict = preprocess(conversations, tokenizer, max_tokens)

        self.input_ids = data_dict["input_ids"]
        self.labels = data_dict["labels"]

    def __len__(self):
        return len(self.input_ids)

    def __getitem__(self, i) -> Dict[str, torch.Tensor]:
        return dict(input_ids=self.input_ids[i], labels=self.labels[i])


@dataclass
class DataCollatorForChatDataset(object):
    """
    Collate examples for supervised fine-tuning.
    """

    tokenizer: transformers.PreTrainedTokenizer

    def __call__(self, instances: Sequence[Dict]) -> Dict[str, torch.Tensor]:
        input_ids, labels = tuple([instance[key] for instance in instances] for key in ("input_ids", "labels"))
        input_ids = torch.nn.utils.rnn.pad_sequence(input_ids, batch_first=True, padding_value=self.tokenizer.pad_token_id)
        labels = torch.nn.utils.rnn.pad_sequence(labels, batch_first=True, padding_value=-100)

        return dict(
            input_ids=input_ids,
            labels=labels,
            attention_mask=input_ids.ne(self.tokenizer.pad_token_id),
        )
    

class ChatDataModule(DataModule):
    def __init__(self, tokenizer: transformers.PreTrainedTokenizer, data_path_train: str, max_tokens = None):

        self.train_dataset = ChatDataset(tokenizer=tokenizer, data_path=data_path_train, max_tokens=max_tokens)
        self.data_collator = DataCollatorForChatDataset(tokenizer=tokenizer)
        

def preprocess(conversations: Sequence[Sequence[dict]], tokenizer: transformers.PreTrainedTokenizer, max_tokens=None) -> Dict:
    """
    Preprocess the data by tokenizing.
    """

    all_input_ids = []
    all_labels = []

    for conv in conversations:
        roles = [msg["role"] for msg in conv]
        messages = [msg["content"] for msg in conv]

        assert roles[0] != "ASSISTANT"
        assert roles[-1] == "ASSISTANT"

        input_messages = []
        if roles[0] == "SYSTEM":
            input_messages.append(SYS_PREFIX+messages[0]+SYS_POSTFIX)

        for role, msg in zip(roles, messages):
            if role == "ASSISTANT":
                input_messages.append(msg + OUTPUT_POSTFIX)
            elif role == "USER":
                input_messages.append(INST_PREFIX + msg + INST_POSTFIX + OUTPUT_PREFIX)

        tokenized_input = tokenizer(input_messages, add_special_tokens=False)

        input_ids = []
        labels = []

        if roles[0] == "SYSTEM":
            input_ids.extend(tokenized_input.input_ids[0])
            labels.extend([-100]*len(tokenized_input.input_ids[0]))

        for role, msg in zip(roles, tokenized_input.input_ids):

            if role == "USER":
                labels.extend([-100]*len(msg))
                input_ids.extend(msg)
            
            elif role == "ASSISTANT":
                labels.extend(msg)
                input_ids.extend(msg)

        if max_tokens is None:
            max_tokens = tokenizer.model_max_length

        input_ids = torch.LongTensor(input_ids)[:max_tokens]
        labels = torch.LongTensor(labels)[:max_tokens]

        assert input_ids.shape == labels.shape   

        all_input_ids.append(input_ids)  
        all_labels.append(labels)


    return dict(input_ids=all_input_ids, labels=all_labels)




from setuptools import setup, find_packages

setup(
    name='llamatune',
    version='0.4.0',
    author='Haven Technologies Inc.',
    author_email='hello@havenllm.com',
    description="Haven\'s Tuning Library for LLM finetuning",
    packages=find_packages(),
    install_requires=[
        'torch==2.0.1',
        'bitsandbytes==0.40.0',
        'einops==0.6.1',
        'evaluate==0.4.0',
        'scikit-learn==1.2.2',
        'sentencepiece==0.1.99',
        'wandb==0.15.3',
        'accelerate==0.21.0',
        'posthog==2.5.0'
    ],
)




from posthog import Posthog
import os

posthog = Posthog('phc_YpKoFD7smPe4SXRtVyMW766uP9AjUwnuRJ8hh2EJcVv',
                  host='https://eu.posthog.com')


def capture_event(event_name, event_properties):
    if not os.environ.get('EVENT_CAPTURE') == "disable":
        posthog.capture("not_distinct", event_name, event_properties)


from llamatune import ChatTrainer
from transformers import HfArgumentParser
from dataclasses import dataclass, field
from typing import Optional
import torch


@dataclass
class TrainingConfig:
    model_name: str = field(default="meta-llama/Llama-2-7b-hf", metadata={"help": 'Huggingface Name of the model you want to train'})
    data_path: str = field(default="data.json", metadata={"help": 'Path towards your training data'})
    output_dir: str = field(default='./trained_model', metadata={"help": 'The output dir for logs and checkpoints'})
    training_recipe: str = field(default="lora", metadata={"help": "Lora Training or Full Training"})
    optim: str = field(default='paged_adamw_8bit', metadata={"help": 'The optimizer to be used'})
    batch_size: int = field(default=1, metadata={"help": 'The training batch size per GPU. Increase for better speed.'})
    gradient_accumulation_steps: int = field(default=16, metadata={"help": 'How many gradients to accumulate before to perform an optimizer step'})
    n_epochs: int = field(default=3, metadata={"help": 'How many optimizer update steps to take'})
    weight_decay: float = field(default=0.0, metadata={"help": 'The L2 weight decay rate of AdamW'}) 
    learning_rate: float = field(default=1e-4, metadata={"help": 'The learning rate'})
    max_grad_norm: float = field(default=0.3, metadata={"help": 'Gradient clipping max norm. This is tuned and works well for all models tested.'})
    gradient_checkpointing: bool = field(default=True, metadata={"help": 'Use gradient checkpointing. You want to use this.'})
    do_train: bool = field(default=True, metadata={"help": 'To train or not to train, that is the question?'})
    lr_scheduler_type: str = field(default='cosine', metadata={"help": 'Learning rate schedule. Constant a bit better than cosine, and has advantage for analysis'})
    warmup_ratio: float = field(default=0.03, metadata={"help": 'Fraction of steps to do a warmup for'})
    logging_steps: int = field(default=1, metadata={"help": 'The frequency of update steps after which to log the loss'})
    group_by_length: bool = field(default=True, metadata={"help": 'Group sequences into batches with same length. Saves memory and speeds up training considerably.'})
    save_strategy: str = field(default='epoch', metadata={"help": 'When to save checkpoints'})
    save_total_limit: int = field(default=3, metadata={"help": 'How many checkpoints to save before the oldest is overwritten'})
    fp16: bool = field(default=True, metadata={"help": 'Whether to use fp16 mixed precision training'})
    tokenizer_type: str = field(default="llama", metadata={"help": "Tokenizer type. Should be \"llama\" for llama models to address tokenizer issue"})
    trust_remote_code: str = field(default=False, metadata={"help": "Whether to trust remote code."})
    compute_dtype: torch.dtype = field(default=torch.float16, metadata={"help":"Compute Datatype for models, either float16 or float32."})
    max_tokens: int = field(default=4096, metadata={"help":"Max tokens"})
    do_eval: bool = field(default=True, metadata={"help": "Whether to evaluate or not"})
    evaluation_strategy: str = field(default="epoch", metadata={"help": "When to evaluate, after certain number of steps or each epoch"})
    use_auth_token: str = field(default=None, metadata={"help": "auth token"})
    use_fast: bool = field(default=False, metadata={"help": "Whether to use fast tokenizer"})
    bits: Optional[int] = field(default=4, metadata={"help": "Number of bits to quantize the model to"})
    double_quant: bool = field(default=True, metadata={"help": "Compress the quantization statistics through double quantization."})
    quant_type: str = field(default="nf4", metadata={"help": "Quantization data type to use. Should be one of `fp4` or `nf4`."})
    lora_r: int = field(default=64, metadata={"help": "Lora R dimension."})
    lora_alpha: float = field(default=16, metadata={"help": " Lora alpha."})
    lora_dropout: float = field(default=0.0, metadata={"help":"Lora dropout."})


hfparser = HfArgumentParser((TrainingConfig))
args = hfparser.parse_args_into_dataclasses(return_remaining_strings=True)[0]

trainer = ChatTrainer(training_config=args)
trainer.train()


import torch

def _get_compute_dtype():
    major, minor = torch.cuda.get_device_capability()
    if major >= 8:
        return torch.bfloat16
    else:
        return torch.float16


def print_trainable_parameters(model):
    """
    Prints the number of trainable parameters in the model.
    """
    trainable_params = 0
    all_param = 0
    for _, param in model.named_parameters():
        all_param += param.numel()
        if param.requires_grad:
            trainable_params += param.numel()
    print(
        f"trainable params: {trainable_params} || "
        f"all params: {all_param} || "
        f"trainable: {100 * trainable_params / all_param}"
    )
